# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import nb_exceptions


def get_gateway_lrp(nat):
    try:
        return nat.gateway_port[0]
    except IndexError:
        raise nb_exceptions.NATNotFound("Port %s has no NAT entry" % nat.uuid)


def get_chassis_hosting_crlrp(nat):
    gateway_lrp = get_gateway_lrp(nat)
    try:
        return gateway_lrp.status[constants.OVN_STATUS_CHASSIS]
    except KeyError:
        raise nb_exceptions.ChassisNotFound(
            "Gateway port %s has no chassis set in its status column.",
            gateway_lrp)
